use std::default::Default;
use std::str::FromStr;

use html5ever::tokenizer::{BufferQueue, TokenizerResult};
use html5ever::tokenizer::{StartTag, TagToken};
use html5ever::tokenizer::{Token, TokenSink, TokenSinkResult, Tokenizer, TokenizerOpts};
use html5ever::{local_name, tendril::*};
use tracing::{debug, error, instrument, trace, warn};
use url::Url;

use crate::db::Website;

impl TokenSink for Website {
    type Handle = Vec<Website>;

    #[instrument(skip(token, _line_number))]
    fn process_token(&self, token: Token, _line_number: u64) -> TokenSinkResult<Self::Handle> {
        match token {
            TagToken(tag) => {
                if tag.kind == StartTag {
                    match tag.name {
                        // this should be all the html
                        // elements that have links
                        local_name!("a")
                        | local_name!("audio")
                        | local_name!("area")
                        | local_name!("img")
                        | local_name!("link")
                        | local_name!("object")
                        | local_name!("source")
                        | local_name!("base")
                        | local_name!("video") => {
                            let mut links = Vec::new();
                            for attr in &tag.attrs {
                                let attr_name = attr.name.local.to_string();
                                if attr_name == "src" || attr_name == "href" || attr_name == "data"
                                {
                                    let url: Option<Url> = match Url::parse(&attr.value) {
                                        Ok(ok) => {
                                            trace!("Found `{}` in the html on `{}` tag", ok.to_string(), tag.name);
                                            Some(ok)
                                        },
                                        Err(e) => {
                                            if attr.value.starts_with('#') {
                                                trace!("Rejecting # url");
                                                None
                                            } else {
                                                match e {
                                                    url::ParseError::RelativeUrlWithoutBase => {
                                                        let origin = self.site.origin().ascii_serialization();
                                                        let url = origin.clone() + &attr.value;
                                                        trace!("Built `{url}` from `{origin} + {}`", &attr.value.to_string());
                                                        if let Ok(url) = Url::parse(&url) {
                                                            trace!("Saved relative url `{}` AS: `{}`", &attr.value, url);
                                                            Some(url)
                                                        } else {
                                                            error!("Failed to reconstruct a url from relative url: `{}` on site: `{}`", &attr.value, self.site.to_string());
                                                            None
                                                        }
                                                    }, 
                                                    _ => {
                                                        error!("MISC error: {:?} {:?}", e, &attr.value);
                                                        None
                                                    },
                                                }
                                            }
                                        },
                                    };
                                    if let Some(mut parsed) = url {
                                        parsed.set_query(None);
                                        parsed.set_fragment(None);
                                        debug!("Final cleaned URL: `{}`", parsed.to_string());
                                        let web = Website::new(&parsed.to_string(), false);
                                        links.push(web);
                                    }
                                }
                            }
                            return TokenSinkResult::Script(links);
                        }
                        local_name!("button") | local_name!("meta") | local_name!("iframe") => {
                            // dbg!(attrs);
                        }
                        _ => {}
                    }
                }
            }
            _ => {}
        }
        TokenSinkResult::Continue
    }
}

#[instrument(skip_all)]
/// Parses the passed site and returns all the sites it links to.
pub async fn parse(site: &Website, data: &str) -> Vec<Website> {
    // prep work
    let mut other_sites: Vec<Website> = Vec::new();

    // change data into something that can be tokenized
    let chunk = Tendril::from_str(data).expect("Failed to parse string into Tendril!");
    // create buffer of tokens and push our input into it
    let token_buffer = BufferQueue::default();
    token_buffer.push_back(
        chunk
            .try_reinterpret::<fmt::UTF8>()
            .expect("Failed to reinterpret chunk!"),
    );
    // create the tokenizer
    let tokenizer = Tokenizer::new(site.clone(), TokenizerOpts::default());

    // go thru buffer
    while let TokenizerResult::Script(mut sites) = tokenizer.feed(&token_buffer) {
        other_sites.append(&mut sites);
        // other_sites.push(sites);
    }
    assert!(token_buffer.is_empty());
    tokenizer.end();

    other_sites
}

